@@ -1,6 +1,6 @@
 -- | This module implements some analyses of Core expressions necessary for
 -- "Test.Inspection". Normally, users of this package can ignore this module.
-{-# LANGUAGE CPP, FlexibleContexts #-}
+{-# LANGUAGE CPP, FlexibleContexts, PatternSynonyms #-}
 module Test.Inspection.Core
   ( slice
   , pprSlice
@@ -12,6 +12,22 @@ module Test.Inspection.Core
   , doesNotContainTypeClasses
   ) where
 
+#if MIN_VERSION_ghc(9,0,0)
+import GHC.Core
+import GHC.Core.Utils
+import GHC.Core.TyCo.Rep
+import GHC.Core.Type
+import GHC.Types.Var as Var
+import GHC.Types.Id
+import GHC.Types.Name
+import GHC.Types.Var.Env
+import GHC.Utils.Outputable as Outputable
+import GHC.Core.Ppr
+import GHC.Core.Coercion
+import GHC.Utils.Misc
+import GHC.Core.DataCon
+import GHC.Core.TyCon (TyCon, isClassTyCon)
+#else
 import CoreSyn
 import CoreUtils
 import TyCoRep
@@ -26,6 +42,11 @@ import Coercion
 import Util
 import DataCon
 import TyCon (TyCon, isClassTyCon)
+#endif
+
+#if MIN_VERSION_ghc(9,2,0)
+import GHC.Types.Tickish
+#endif
 
 import qualified Data.Set as S
 import Control.Monad.State.Strict
@@ -33,6 +54,12 @@ import Control.Monad.Trans.Maybe
 import Data.List (nub)
 import Data.Maybe
 
+#if !MIN_VERSION_ghc(9,2,0)
+pattern Alt :: a -> b -> c -> (a, b, c)
+pattern Alt a b c = (a, b, c)
+{-# COMPLETE Alt #-}
+#endif
+
 type Slice = [(Var, CoreExpr)]
 
 -- | Selects those bindings that define the given variable (with this variable first)
@@ -50,7 +77,7 @@ slice binds v
         seen <- gets (v `S.member`)
         unless seen $ do
             modify (S.insert v)
-            let Just e = lookup v binds
+            let e = fromJust $ lookup v binds
             go e
           | otherwise = return ()
 
@@ -66,7 +93,7 @@ slice binds v
     go (Type _)                    = pure ()
     go (Coercion _)                = pure ()
 
-    goA (_, _, e) = go e
+    goA (Alt _ _ e) = go e
 
 -- | Pretty-print a slice
 pprSlice :: Slice -> SDoc
@@ -76,15 +103,15 @@ pprSlice slice =
 -- | Pretty-print two slices, after removing variables occurring in both
 pprSliceDifference :: Slice -> Slice -> SDoc
 pprSliceDifference slice1 slice2 =
-    nest 4 (hang (text "LHS" Outputable.<> colon) 4 (pprSlice slice1')) $$
-    nest 4 (hang (text "RHS" Outputable.<> colon) 4 (pprSlice slice2'))
+    hang (text "LHS" Outputable.<> colon) 4 (pprSlice slice1') $$
+    hang (text "RHS" Outputable.<> colon) 4 (pprSlice slice2')
   where
     both = S.intersection (S.fromList (map fst slice1)) (S.fromList (map fst slice2))
     slice1' = filter (\(v,_) -> v `S.notMember` both) slice1
     slice2' = filter (\(v,_) -> v `S.notMember` both) slice2
 
 withLessDetail :: SDoc -> SDoc
-#if MIN_VERSION_GLASGOW_HASKELL(8,2,0,0)
+#if MIN_VERSION_GLASGOW_HASKELL(8,2,0,0) && !MIN_VERSION_GLASGOW_HASKELL(9,0,0,0)
 withLessDetail sdoc = sdocWithDynFlags $ \dflags ->
      withPprStyle (defaultUserStyle dflags) sdoc
 #else
@@ -98,7 +125,7 @@ type VarPairSet = S.Set VarPair
 -- have auxiliary variables in the right order.
 -- (This is mostly to work-around the buggy CSE in GHC-8.0)
 -- It also breaks if there is shadowing.
-eqSlice :: Bool {- ^ ignore types -} -> Slice -> Slice -> Bool
+eqSlice :: Bool {- ^ ignore types and hpc ticks -} -> Slice -> Slice -> Bool
 eqSlice _ slice1 slice2 | null slice1 || null slice2 = null slice1 == null slice2
   -- Mostly defensive programming (slices should not be empty)
 eqSlice it slice1 slice2
@@ -138,7 +165,11 @@ eqSlice it slice1 slice2
     essentiallyVar (App e a)  | it, isTyCoArg a = essentiallyVar e
     essentiallyVar (Lam v e)  | it, isTyCoVar v = essentiallyVar e
     essentiallyVar (Cast e _) | it              = essentiallyVar e
+#if MIN_VERSION_ghc(9,0,0)
+    essentiallyVar (Case s _ _ [Alt _ _ e]) | it, isUnsafeEqualityProof s = essentiallyVar e
+#endif
     essentiallyVar (Var v)                      = Just v
+    essentiallyVar (Tick HpcTick{} e) | it      = essentiallyVar e
     essentiallyVar _                            = Nothing
 
     go :: RnEnv2 -> CoreExpr -> CoreExpr -> MaybeT (State (S.Set (Var,Var))) ()
@@ -150,12 +181,18 @@ eqSlice it slice1 slice2
 
     go env (Cast e1 _) e2 | it             = go env e1 e2
     go env e1 (Cast e2 _) | it             = go env e1 e2
+#if MIN_VERSION_ghc(9,0,0)
+    go env (Case s _ _ [Alt _ _ e1]) e2 | it, isUnsafeEqualityProof s = go env e1 e2
+    go env e1 (Case s _ _ [Alt _ _ e2]) | it, isUnsafeEqualityProof s = go env e1 e2
+#endif
     go env (Cast e1 co1) (Cast e2 co2)     = do guard (eqCoercionX env co1 co2)
                                                 go env e1 e2
 
     go env (App e1 a) e2 | it, isTyCoArg a = go env e1 e2
     go env e1 (App e2 a) | it, isTyCoArg a = go env e1 e2
     go env (App f1 a1)   (App f2 a2)       = go env f1 f2 >> go env a1 a2
+    go env (Tick HpcTick{} e1) e2 | it     = go env e1 e2
+    go env e1 (Tick HpcTick{} e2) | it     = go env e1 e2
     go env (Tick n1 e1)  (Tick n2 e2)      = guard (go_tick env n1 n2) >> go env e1 e2
 
     go env (Lam b e1) e2 | it, isTyCoVar b = go env e1 e2
@@ -191,11 +228,16 @@ eqSlice it slice1 slice2
     go _ _ _ = guard False
 
     -----------
-    go_alt env (c1, bs1, e1) (c2, bs2, e2)
+    go_alt env (Alt c1 bs1 e1) (Alt c2 bs2 e2)
       = guard (c1 == c2) >> go (rnBndrs2 env bs1 bs2) e1 e2
 
+#if MIN_VERSION_ghc(9,2,0)
+    go_tick :: RnEnv2 -> CoreTickish -> CoreTickish -> Bool
+    go_tick env (Breakpoint _ lid lids) (Breakpoint _ rid rids)
+#else
     go_tick :: RnEnv2 -> Tickish Id -> Tickish Id -> Bool
     go_tick env (Breakpoint lid lids) (Breakpoint rid rids)
+#endif
           = lid == rid  &&  map (rnOccL env) lids == map (rnOccR env) rids
     go_tick _ l r = l == r
 
@@ -230,7 +272,7 @@ allTyCons ignore slice =
 
     goB (b, e) = goV b ++ go e
 
-    goA (_,pats, e) = concatMap goV pats ++ go e
+    goA (Alt _ pats e) = concatMap goV pats ++ go e
 
     goT (TyVarTy _)      = []
     goT (AppTy t1 t2)    = goT t1 ++ goT t2
@@ -239,6 +281,9 @@ allTyCons ignore slice =
     goT (ForAllTy _ t)   = goT t
 #if MIN_VERSION_GLASGOW_HASKELL(8,2,0,0)
     goT (FunTy
+#if MIN_VERSION_GLASGOW_HASKELL(9,0,0,0)
+               _
+#endif
 # if MIN_VERSION_GLASGOW_HASKELL(8,9,0,0)
                _
 # endif
@@ -273,7 +318,7 @@ freeOfTerm slice needles = listToMaybe [ (v,e) | (v,e) <- slice, not (go e) ]
 
     goB (_, e) = go e
 
-    goA (ac, _, e) = goAltCon ac && go e
+    goA (Alt ac _ e) = goAltCon ac && go e
 
     goAltCon (DataAlt dc) | isNeedle (dataConName dc) = False
     goAltCon _ = True
@@ -320,7 +365,7 @@ doesNotAllocate slice = listToMaybe [ (v,e) | (v,e) <- slice, not (go (idArity v
         -- A let binding allocates if any variable is not a join point and not
         -- unlifted
 
-    goA a (_,_, e) = go a e
+    goA a (Alt _ _ e) = go a e
 
 doesNotContainTypeClasses :: Slice -> [Name] -> Maybe (Var, CoreExpr, [TyCon])
 doesNotContainTypeClasses slice tcNs
