@@ -5,7 +5,12 @@
 {-# LANGUAGE MultiWayIf #-}
 {-# LANGUAGE TemplateHaskell #-}
 {-# LANGUAGE LambdaCase #-}
-module Test.Inspection.Plugin (plugin) where
+module Test.Inspection.Plugin
+  ( plugin
+  , checkProperty
+  , CheckResult(..)
+  , prettyProperty
+  ) where
 
 import Control.Monad
 import System.Exit
@@ -16,17 +21,39 @@ import Data.List
 import qualified Data.Map.Strict as M
 import qualified Language.Haskell.TH.Syntax as TH
 
+#if MIN_VERSION_ghc(9,4,0)
+import GHC.Types.Error
+import GHC.Driver.Session
+#endif
+
+#if MIN_VERSION_ghc(9,0,0)
+import GHC.Plugins hiding (SrcLoc)
+import GHC.Utils.Outputable as Outputable
+#else
 import GhcPlugins hiding (SrcLoc)
 import Outputable
+#endif
+
+#if MIN_VERSION_ghc(9,2,0)
+import GHC.Types.TyThing
+#endif
 
 import Test.Inspection (Obligation(..), Property(..), Result(..))
 import Test.Inspection.Core
 
 -- | The plugin. It supports some options:
 --
--- * @-fplugin-opt=Test.Inspection.Plugin:keep-going@ to ignore a failing build
--- * @-fplugin-opt=Test.Inspection.Plugin:keep-going-O0@ to ignore a failing build when optimisations are off
+-- * @-fplugin-opt=Test.Inspection.Plugin:keep-going@ to keep building despite failing obligations
+-- * @-fplugin-opt=Test.Inspection.Plugin:keep-going-O0@ to keep building despite failing obligations, when optimisations are off
+-- * @-fplugin-opt=Test.Inspection.Plugin:skip-O0@ to skip performing inspections when optimisations are off
 -- * @-fplugin-opt=Test.Inspection.Plugin:quiet@ to be silent if all obligations are fulfilled
+--
+-- It makes sense to enable only one of @keep-going@, @keep-going-O0@ and
+-- @skip-O0@ at a time. @skip-O0@ is useful when working with GHCi, to suppress
+-- inspection failure messages and eliminate the overhead of inspection when
+-- loading.
+
+
 plugin :: Plugin
 plugin = defaultPlugin
     { installCoreToDos = install
@@ -35,7 +62,7 @@ plugin = defaultPlugin
 #endif
     }
 
-data UponFailure = AbortCompilation | KeepGoingO0 | KeepGoing deriving Eq
+data UponFailure = AbortCompilation | KeepGoingO0 | SkipO0 | KeepGoing deriving Eq
 
 data ReportingMode = Verbose | Quiet deriving Eq
 
@@ -47,6 +74,7 @@ install args passes = return $ passes ++ [pass]
     pass = CoreDoPluginPass "Test.Inspection.Plugin" (proofPass upon_failure report)
     upon_failure | "keep-going" `elem` args    = KeepGoing
                  | "keep-going-O0" `elem` args = KeepGoingO0
+                 | "skip-O0" `elem` args       = SkipO0
                  | otherwise                   = AbortCompilation
     report | "quiet" `elem` args = Quiet
            | otherwise           = Verbose
@@ -73,18 +101,19 @@ prettyObligation mod (Obligation {..}) result =
   where
     name = case testName of
         Just n -> n
-        Nothing -> prettyProperty mod target property
+        Nothing -> prettyProperty (showTHName mod) target property
 
-prettyProperty :: Module -> TH.Name -> Property -> String
-prettyProperty mod target (EqualTo n2 False)  = showTHName mod target ++ " === " ++ showTHName mod n2
-prettyProperty mod target (EqualTo n2 True)   = showTHName mod target ++ " ==- " ++ showTHName mod n2
-prettyProperty mod target (NoTypes [t])       = showTHName mod target ++ " `hasNoType` " ++ showTHName mod t
-prettyProperty mod target (NoTypes ts)        = showTHName mod target ++ " mentions none of " ++ intercalate ", " (map (showTHName mod) ts)
-prettyProperty mod target NoAllocation        = showTHName mod target ++ " does not allocate"
-prettyProperty mod target (NoTypeClasses [])  = showTHName mod target ++ " does not contain dictionary values"
-prettyProperty mod target (NoTypeClasses ts)  = showTHName mod target ++ " does not contain dictionary values except of " ++ intercalate ", " (map (showTHName mod) ts)
-prettyProperty mod target (NoUseOf ns)        = showTHName mod target ++ " uses none of " ++ intercalate ", " (map (showTHName mod) ns)
-prettyProperty mod target CoreOf              = showTHName mod target ++ " core dump" -- :)
+prettyProperty :: (TH.Name -> String) -> TH.Name -> Property -> String
+prettyProperty showName target = \case
+  EqualTo n2 False -> showName target ++ " === " ++ showName n2
+  EqualTo n2 True  -> showName target ++ " ==- " ++ showName n2
+  NoTypes [t]      -> showName target ++ " `hasNoType` " ++ showName t
+  NoTypes ts       -> showName target ++ " mentions none of " ++ intercalate ", " (map showName ts)
+  NoAllocation     -> showName target ++ " does not allocate"
+  NoTypeClasses [] -> showName target ++ " does not contain dictionary values"
+  NoTypeClasses ts -> showName target ++ " does not contain dictionary values except of " ++ intercalate ", " (map showName ts)
+  NoUseOf ns       -> showName target ++ " uses none of " ++ intercalate ", " (map showName ns)
+  CoreOf           -> showName target ++ " core dump" -- :)
 
 -- | Like show, but omit the module name if it is he current module
 showTHName :: Module -> TH.Name -> String
@@ -293,42 +322,35 @@ resultToExpr (Failure s) = App <$> dcExpr 'Failure <*> mkStringExpr s
 
 proofPass :: UponFailure -> ReportingMode -> ModGuts -> CoreM ModGuts
 proofPass upon_failure report guts = do
-    dflags <- getDynFlags
-    let noopt = optLevel dflags < 1
-    when noopt $
-        warnMsg
-#if MIN_VERSION_GLASGOW_HASKELL(8,9,0,0)
-                NoReason
-#endif
-                $ fsep $ map text $ words "Test.Inspection: Compilation without -O detected. Expect optimizations to fail."
+    case upon_failure of
+        SkipO0 -> pure guts
+        _ -> do
+            let (guts', obligations) = extractObligations guts
+            (toStore, stats) <- (concat `bimap` M.unionsWith (+)) . unzip <$>
+                mapM (checkObligation report guts') obligations
+            let n = sum stats :: Int
 
-    let (guts', obligations) = extractObligations guts
-    (toStore, stats) <- (concat `bimap` M.unionsWith (+)) . unzip <$>
-        mapM (checkObligation report guts') obligations
-    let n = sum stats :: Int
+            guts'' <- storeResults toStore  guts'
 
-    guts'' <- storeResults toStore  guts'
+            let q :: Stat -> Int
+                q s = fromMaybe 0 $ M.lookup s stats
 
-    let q :: Stat -> Int
-        q s = fromMaybe 0 $ M.lookup s stats
+            let summary_message = nest 2 $
+                    vcat [ nest 2 (desc s) Outputable.<> colon <+> ppr (q s)
+                         | s <- [minBound..maxBound], q s > 0 ]
 
-    let summary_message = nest 2 $
-            vcat [ nest 2 (desc s) Outputable.<> colon <+> ppr (q s)
-                 | s <- [minBound..maxBound], q s > 0 ]
+            -- Only print a message if there are some compile-time results to report
+            unless (q StoredResult == n) $ do
+                if q ExpSuccess + q ExpFailure + q StoredResult == n
+                then unless (report == Quiet) $
+                        putMsg $ text "inspection testing successful" $$ summary_message
+                else do
+                    errorMsg $ text "inspection testing unsuccessful" $$ summary_message
+                    case upon_failure of
+                        KeepGoing           -> return ()
+                        _                   -> liftIO $ exitFailure -- kill the compiler. Is there a nicer way?
 
-    -- Only print a message if there are some compile-time results to report
-    unless (q StoredResult == n) $ do
-        if q ExpSuccess + q ExpFailure + q StoredResult == n
-        then unless (report == Quiet) $
-                putMsg $ text "inspection testing successful" $$ summary_message
-        else do
-            errorMsg $ text "inspection testing unsuccessful" $$ summary_message
-            case upon_failure of
-                KeepGoing           -> return ()
-                KeepGoingO0 | noopt -> return ()
-                _                   -> liftIO $ exitFailure -- kill the compiler. Is there a nicer way?
-
-    return guts''
+            return guts''
 
 
 desc :: Stat -> SDoc
